{
   umod, mod functionality
   Copyright (c) 2011. Dejan Boras

   This file is part of OpenArena Name Changer (oanc).

   OpenArena Name Changer (oanc) is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 2 of the License, or
   (at your option) any later version.

   OpenArena Name Changer (oanc) is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with OpenArena Name Changer (oanc). If not, see <http://www.gnu.org/licenses/>.

   Started On:    15.04.2011.
   Last Update:   25.08.2011.
}

{$MODE OBJFPC}{$H+}
UNIT umod;

INTERFACE

USES
   Dialogs,
   uoanc, frmuModSelect, frmuFav;

VAR
   retModPlayerName: string;

procedure modDlgSaveOpen();
procedure modDlgLoadOpen();

function modSave(): boolean;
function modLoad(): boolean;

function singleModSave(): boolean;
function singleModLoad(): boolean;

IMPLEMENTATION

{load the player name into a list for all mods}
procedure loadModNameList();
var
   modname: string;
   i: longint;

begin
   namesList.Clear();

   if(oaHasMods()) then begin
      for i := 0 to (modlist.Count-1) do begin
         modname := oaConfigLoad(modlist.Strings[i]);
         namesList.Add(modname);
      end;
   end;
end;

procedure modDlgSaveOpen();
begin
   {prepare form}
   frmModSelect.btnAction.Caption := 'Save to checked mods';
   frmModSelect.lblChoose.Caption := 'Choose mods to save your name to';
   frmModSelect.Caption := 'Player Name Save';
   frmModSelect.btnSingleItemAction.Caption := 'Save name to highlighted mod';

   {set callback}
   SelectActionFunc := @modSave;
   SingleItemActionFunc := @singleModSave;

   {load list of names}
   loadModNameList();

   {open the mod selection dialog}
   frmModSelect.ShowModal();
end;

procedure modDlgLoadOpen();
begin
   retModPlayerName := '';

   {prepare form}
   frmModSelect.btnAction.Caption := 'Add Checked Mods to Favorites';
   frmModSelect.lblChoose.Caption := 'Choose mods to load your name from';
   frmModSelect.Caption := 'Player Name Load';
   frmModSelect.btnSingleItemAction.Caption := 'Load name from highlighted mod';

   {set callback}
   SelectActionFunc := @modLoad;
   SingleItemActionFunc := @singleModLoad;

   {load list of names}
   loadModNameList();

   {open the mod selection dialog}
   frmModSelect.ShowModal();
end;

function modSave(): boolean;
var
   i: longint;
   hasMods: boolean;
   save: boolean = false;

begin
   result := false;

   {check if there are mods}
   hasMods := oaHasMods();
   modlistsel.Clear();

   {if there are mods then add them to the mod selection list}
   if(hasMods) then begin
      for i := 0 to (frmModSelect.cblMods.Items.Count-1) do begin
         if(frmModSelect.cblMods.Checked[i]) then
            modlistsel.Add(modlist[i]);
      end;

      if(modlistsel.Count > 0) then save := true;
   end else
      save := true;

   {if anything available for saving the save}
   if(save) then begin
      oaSaveToMods();
      result := true;
   {otherwise give a message}
   end else
      MessageDlg('Error', 'No mods are selected.', mtWarning, [mbOK], 0);
end;

function modLoad(): boolean;
var
   i: longint;
   hasMods: boolean;
   name: string;
   loaded: boolean = false;

begin
   result := false;

   hasMods := oaHasMods();
   {if there are mods}
   if(hasMods) then begin
      {add selected mod names to the favorites}
      for i := 0 to (frmModSelect.cblMods.Items.Count-1) do begin
         if(frmModSelect.cblMods.Checked[i]) then begin
            name := namesList[i];

            {add only if the name does not already exist}
            if(FindFav(name) < 0) then
                  AddToFav(name);

            loaded := true;
         end;
      end;
   {if there is only config file}
   end else begin
      writeln(frmModSelect.cblMods.Items.Count);
      writeln(frmModSelect.cblMods.Checked[0]);
   end;

   if(loaded) then begin
      MessageDlg('Success', 'Selected mod names added to favorites.', mtInformation, [mbOK], 0);
      result := true;
   end else
      MessageDlg('Sorry', 'You need to select some mods to be able to load them into favorites.', mtWarning, [mbOK], 0);
end;

function singleModSave(): boolean;
var
   hasMods: boolean;
   modname: string;

begin
   result := false;

   hasMods := oaHasMods();

   if(frmModSelect.cblMods.ItemIndex > -1) then begin
      if(hasMods) then begin
         modname := frmModSelect.cblMods.Items[frmModSelect.cblMods.ItemIndex];
         oaSaveMod(modname);
         namesList[frmModSelect.cblMods.ItemIndex] := PlayerName;
         SetPreview(PlayerName);
      end else begin
         modlistsel.Clear();
         oaSaveToMods();
      end;

      result := true;
   end;
end;

function singleModLoad(): boolean;
var
   hasMods: boolean;
   name: string;

begin
   result := false;

   hasMods := oaHasMods();

   if(frmModSelect.cblMods.ItemIndex > -1) then begin
      if(hasMods) then begin
         name := namesList[frmModSelect.cblMods.ItemIndex];
         retModPlayerName := name;
      end else begin
         retModPlayerName := oaConfigLoad(oacfgfile);
      end;

      frmModSelect.Close();
      result := true;
   end;
end;

END.

