{
   uoanc, base unit for OA Name Changer
   Copyright (c) 2011. Dejan Boras

   This file is part of OpenArena Name Changer (oanc).

   OpenArena Name Changer (oanc) is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 2 of the License, or
   (at your option) any later version.

   OpenArena Name Changer (oanc) is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with OpenArena Name Changer (oanc). If not, see <http://www.gnu.org/licenses/>.
   
   Started On:    01.03.2011.
   Last Update:   29.08.2011.
}

{$MODE OBJFPC}{$H+}{$I-}
UNIT uoanc;

INTERFACE

   USES classes, sysutils, Graphics, ExtCtrls, BGRABitmap, BGRABitmapTypes,
   IniFiles, Dialogs, Forms, FileUtil;

CONST
   { characters }
   MAX_COLOR = 8;
   MAX_CHAR = 127;
   CHAR_WIDTH = 16;
   CHAR_HEIGHT = 16;

   HCHARS   = 16;
   VCHARS   = 8;

   DEFAULT_PLAYER_NAME = 'UnnamedPlayer';

{ colors }

TYPE
   TClr = array[0..2] of byte;

CONST
   { OpenArena colors }
   CLR_BLACK         = 0000;
   CLR_RED           = 0001;
   CLR_GREEN         = 0002;
   CLR_YELLOW        = 0003;
   CLR_BLUE          = 0004;
   CLR_AQUA          = 0005;
   CLR_PURPLE        = 0006;
   CLR_WHITE         = 0007;
   CLR_ORANGE        = 0008;

   colors: array[0..MAX_COLOR] of TClr =
    ((000, 000, 000), {black}
     (255, 000, 000), {red}
     (000, 255, 000), {green}
     (255, 255, 000), {yellow}
     (000, 000, 255), {blue}
     (000, 255, 255), {aqua}
     (255, 000, 255), {purple}
     (255, 255, 255), {the racist color}
     (255, 160, 000));{orange}

   { config }

   {configuration defaults}
   cfgFileName  = 'oanc.ini';
   cfgDirectory = '.oanc'+DirectorySeparator;
   lcfgFN = 'writeconfighere';

   oaCFG_FN = 'q3config.cfg';
   oaCFG_BASE_NAME = 'baseoa';

   {returned by oaCheckCfg()}
   oaCFG_CHECK_OK          = 0;
   oaCFG_CHECK_DEFAULT     = 1;
   oaCFG_CHECK_FAIL        = 2;

   {save program configuration locally?}
   localconfig: boolean = false;

TYPE
   {window properties}
   TWindowProps = record
      loaded: boolean;
      w, h, x, y: longint;
   end;

VAR
   {images}
   imgs: array[0..127] of TBGRABitmap;
   charselimg: TBGRABitmap;

   loadedname, {name that was loaded from either oanc config file or oa default config file }
   cfgPlayerName, {name loaded from config file}
   PlayerName, {current player name}
   oaDefaultName: string; {player name}
   oacfgfile: string; {openarean config file}

   {configuration}
   localCfgPath, localCfgFile: string; {local configuration path}
   cfgPath: string; {path to configuration}
   cfgFile: string; {configuration file name}
   iniCfg: TIniFile; {.INI configuration file object}

   modlist: TStringList; {list of mods}
   modlistsel: TStringList; {list of selected mods}

   {callbacks for favorites configuration}
   LoadFavorites, SaveFavorites: TProcedure;

   {window properties, loaded and saved with the configuration file}
   favWndProps, mainWndProps, modSelectWndProps, cfgWndProps: TWindowProps;

{ IMAGES }
{loads character images}
function LoadImages(): boolean;
{creates the character selection image out of individual character images}
procedure CreateCharSelectImage();

{ CONFIGURATION }
{initializes program configuration}
procedure InitializeConfig();
{loads program configuration}
procedure LoadConfig();
{saves program configuration}
procedure SaveConfig();

{returns the file name and path to the OA configuration file}
function oaGetCfgFile(): string;
{checks whether the oa config file exist at the specified location}
function oaCfgFileExists(const fn: string): boolean;
{construct oa filename based on the given modname}
function oaConstructCfgFN(const modfn: string): string;
{check the set oa config}
function oaCheckConfig(default: boolean): longint;
{load the OA config}
function oaConfigLoad(const modfn: string): string;
{save playername to OA config}
function oaConfigSave(const fn: string): boolean;

{MODS}
{saves to a specific mod configuration file}
function oaSaveMod(const modname: string): boolean;
{saves to all mod configuration files}
procedure oaSaveToMods();
{finds mods and gets a list of player names}
procedure oaScanMods();
{checks if there are any mods}
function oaHasMods(): boolean;
{returns the OA base path}
function oaBasePath(): string;

{ RENDERING }
{renders a preview of a name on the specified shape}
procedure PreviewName(const thename: string; const shp: TShape);

{ WINDOW PROPERTIES }
procedure LoadWndProps(const section: string; var props: TWindowProps);
procedure WriteWndProps(const section: string; var props: TWindowProps);
procedure ApplyWndProps(var props: TWindowProps; var frm: TForm);
procedure StoreWndProps(var props: TWindowProps; var f: TForm);

{ URL ENCODING }

function urlEncode(const s: string): string;
function urlDecode(const s: string): string;

IMPLEMENTATION

{ IMAGES }

function LoadImages(): boolean;
var
   i: longint;
   ns: string;

begin
   result := false;

   for i := 1 to 127 do begin
      str(i, ns);
      try
         imgs[i] := TBGRABitmap.Create('chars'+DirectorySeparator+ns+'.png');
      except
         MessageDlg('Error', 'Can not load a character image.', mtError, [mbOK], 0);
         Application.Terminate();
         exit;
      end;
   end;

   result := true;
end;

procedure CreateCharSelectImage();
var
   c, x, y: longint;

begin
   charselimg := TBGRABitmap.Create(HCHARS*CHAR_WIDTH, VCHARS*CHAR_WIDTH);

   for x := 0 to (HCHARS-1) do begin
      for y := 0 to (VCHARS-1) do begin
         c := y*HCHARS+x;
         if(c > 0) and (c <= MAX_CHAR) then
            imgs[c].Draw(charselimg.Canvas, x*CHAR_WIDTH, y*CHAR_HEIGHT, false);
      end;
   end;
end;

{ COFNIGURATION }

procedure InitializeConfig();
begin
   localCfgFile := localCfgPath + cfgFileName;
   if(FileExists(localCfgPath+lcfgFN)) then
      localconfig := true;

   {determine config path and file}
   cfgPath := ExtractFilePath(ChompPathDelim(SysToUTF8(GetAppConfigDir(False)))) + '.oanc';
   cfgFile := cfgPath+DirectorySeparator+cfgFileName;
end;

function getOANCConfigName(): string;
begin
   if(not localconfig) then
      result := cfgFile
   else
      result := localCfgFile
end;

procedure LoadConfig();
var
   fn: string;

begin
   fn := getOANCConfigName();

   iniCfg := TIniFile.Create(Utf8ToSys(fn), true);
   try
      loadedname := urlDecode(iniCfg.ReadString('General', 'Name', DEFAULT_PLAYER_NAME));
      cfgPlayerName := loadedname;
      oacfgfile := SysToUTF8(iniCfg.ReadString('General', 'OAConfigFile', ''));

      LoadWndProps('frmMain', mainWndProps);
      LoadWndProps('frmFav', favWndProps);
      LoadWndProps('frmModSelect', modSelectWndProps);
      LoadWndProps('frmCfg', cfgWndProps);

      if(LoadFavorites <> nil) then
         LoadFavorites();
   except
   end;

   if(iniCfg <> nil) then
      iniCfg.Free();
end;

procedure SaveConfig();
var
   fn: string;

begin
   if(not localconfig) then
      CreateDirUTF8(cfgPath);

   fn := getOANCConfigName();

   try
      iniCfg := TIniFile.Create(Utf8ToSys(fn), true);

      iniCfg.WriteString('General', 'OAConfigFile', UTF8ToSys(oacfgfile));

      WriteWndProps('frmMain', mainWndProps);
      WriteWndProps('frmFav', favWndProps);
      WriteWndProps('frmModSelect', modSelectWndProps);
      WriteWndProps('frmCfg', cfgWndProps);

      if(SaveFavorites <> nil) then
         SaveFavorites();

      iniCfg.WriteString('General', 'Name', urlEncode(playername));
   except
   end;

   if(iniCfg <> nil) then
      iniCfg.Free();
end;

function oaGetCfgFile(): string;
var
   path: string;

begin
   path := '';
   {$IFDEF WINDOWS} {MS-WINDOWS}
   path := IncludeTrailingPathDelimiter(GetEnvironmentVariableUTF8('APPDATA')) +
      'OpenArena' + DirectorySeparator +
      'baseoa' + DirectorySeparator + oaCFG_FN;
   {$ENDIF}
   {$IFDEF UNIX} {LINUX and UNICES}
   path := IncludeTrailingPathDelimiter(GetEnvironmentVariableUTF8('HOME')) +
      '.openarena' + DirectorySeparator +
      'baseoa' + DirectorySeparator + oaCFG_FN;
   {$ENDIF}
   {$IFDEF DARWIN} {MACOSX}
   path := IncludeTrailingPathDelimiter(GetEnvironmentVariableUTF8('HOME')) +
      '.openarena' + DirectorySeparator +
      'baseoa' + DirectorySeparator + oaCFG_FN;
   {$ENDIF}

   result := path;
end;

function oaCfgFileExists(const fn: string): boolean;
begin
   result := FileExistsUTF8(fn);
end;

function oaConstructCfgFN(const modfn: string): string;
var
   path: string;

begin
   path := ExcludeTrailingPathDelimiter(oaBasePath());
   result := path + DirectorySeparator+modfn+DirectorySeparator+oaCFG_FN;
end;

function oaCheckConfig(default: boolean): longint;
var
   exists, defaultexists: boolean;

procedure setDefault();
begin
   oacfgfile := oaGetCfgFile();

   defaultexists := oaCfgFileExists(oacfgfile);

   if(defaultexists) then
      result := oaCFG_CHECK_OK
   else begin
      MessageDlg('Warning', 'No OA config file found.', mtWarning, [mbOK], 0);
      result := oaCFG_CHECK_FAIL;
   end;
end;

begin
   if(default) then begin
      setDefault(); exit;
   end;

   exists := oaCfgFileExists(oacfgfile);
   if(not exists) then begin
      setDefault();

      if(defaultexists) then begin
         MessageDlg('Warning', 'The previously selected OpenArena config file was not found, '+
            'so default one was selected.', mtWarning, [mbOK], 0);
      end;
   end else
       result := oaCFG_CHECK_OK;

   oaScanMods();
end;

{save playername to OA config}
function oaConfigLoad(const modfn: string): string;
var
   f: text;
   l: string;
   buf1: array[0..1023] of byte;
   err: longint;

begin
   result := '';

   {open the oa config file}
   Assign(f, UTF8ToSys(oaConstructCfgFN(modfn)));
   SetTextBuf(f, buf1, 1024);

   Reset(f);
   err := IOResult();
   if(err = 0) then begin
      {go through the oa cfg file}
      repeat
         readln(f, l);
         err := IOResult();
         if(err = 0) then begin
            if(pos('seta name', l) <> 0) then begin
               {get the player name}
               delete(l, 1, Length('seta name "'));
               result := copy(l, 1, Length(l)-1);
               break;
            end;
            err := IOResult();
            if(err <> 0) then break;
         end else break;
      until eof(f);

      {close both files}
      Close(f);
      IOResult();
   end;
end;

function oaConfigSave(const fn: string): boolean;
const
   errtitle = 'OA Config File Error';
   nochgstr = 'No changes were made.';

var
   f, nf: text;
   l, nfn, bakfn: string;
   buf1, buf2: array[0..1023] of byte;
   err: longint;

begin
   result := false;

   {create a new file}
   nfn := IncludeTrailingPathDelimiter(ExtractFilePath(fn))+'temp.new';
   Assign(nf, UTF8ToSys(nfn));
   Rewrite(nf);
   SetTextBuf(nf, buf1, 1024);

   {open the oa config file}
   Assign(f, UTF8ToSys(fn));
   SetTextBuf(f, buf2, 1024);

   Reset(f);
   err := IOResult();
   if(err = 0) then begin

      {go through the oa cfg file}
      repeat
         readln(f, l);
         err := IOResult();
         if(err = 0) then begin
            if(pos('seta name', l) <> 0) then begin
               writeln(nf, 'seta name "'+playername+'"');
            end else begin
               writeln(nf, l);
            end;
            err := IOResult();
            if(err <> 0) then begin
               MessageDlg(errtitle, 'A new file could not be written for: '+fn+'. '+nochgstr,
                  mtError, [mbOK], 0);
               break;
            end;
         end else begin
            MessageDlg(errtitle, ' The config file could not be read.'+nochgstr,
               mtError, [mbOK], 0);
            break;
         end;
      until eof(f);

      {close both files}
      Close(f);
      err := IOResult();
      if(err <> 0) then begin
         MessageDlg(errtitle, 'The config file: '+fn+' could not be closed.'+nochgstr,
            mtError, [mbOK], 0);
      end;

      Close(nf);
      err := IOResult();
      if(err <> 0) then begin
         MessageDlg(errtitle, 'The new OA config file could not be closed for: '+fn+'. '+nochgstr,
            mtError, [mbOK], 0);
         exit;
      end;

      {get the name of the backup file}
      bakfn := IncludeTrailingPathDelimiter(ExtractFilePath(fn))+DirectorySeparator+ExtractFilename(fn)+'.oancbak';

      {erase the old backup file}
      assign(f, UTF8ToSys(bakfn));
      erase(f);
      IOResult();

      {rename the oa config file to backup file}
      assign(f, UTF8ToSys(fn));
      rename(f, UTF8ToSys(bakfn));
      IOResult();

      {rename the new file to oa config file}
      rename(nf, UTF8ToSys(fn));
      err := IOResult();
      if(IOResult() <> 0) then begin
         MessageDlg(errtitle, 'Could not set the new file for: '+fn+' as OA config file. The old file was preserved as: '+bakfn,
            mtError, [mbOK], 0);
         exit;
      end;
   end else
      MessageDlg(errtitle, ' The file: '+fn+' cannot be opened.'+#10+
      'Please set a valid config file in OANC configuration options.', mtError, [mbOK], 0);

   if(err = 0) then
      result := true;
end;

{MODS}
function oaSaveMod(const modname: string): boolean;
var
   cf, path: string;

begin
   path := ExcludeTrailingPathDelimiter(oaBasePath());
   cf := path+DirectorySeparator+modname+DirectorySeparator+oaCFG_FN;

   result := oaConfigSave(cf);
end;

procedure oaSaveToMods();
var
   i: longint;
   err: boolean = false; {indicates if an error condition was encountered}
   errs: string = ''; {contains a list of files for which there were errors}

begin
   {if mods were selected}
   if(modlistsel.Count > 0) then begin
      for i := 0 to (modlistsel.Count-1) do begin
         if(modlistsel[i] = 'baseoa') then
            oaDefaultName := PlayerName;

         {save the config file}
         if(oaSaveMod(modlistsel[i]) = false) then begin
            err := true;
            errs := errs + modlistsel[i] + #10;
         end;
      end;

      {message}
      if(not err) then
         MessageDlg('Success', 'All config files saved successfully.', mtInformation, [mbOK], 0)
      else
         MessageDlg('Error', 'Error saving config files: '+#10+errs, mtError, [mbOK], 0);
   {if no mods then try to write directly to config file}
   end else begin
      if(oaConfigSave(oacfgfile)) then
         MessageDlg('Success', 'OpenArena configuration file saved successfully.', mtInformation, [mbOK], 0);
   end;
end;

procedure oaScanMods();
var
   src: TSearchRec;
   result: longint;
   path, cf: string;

begin
   {we need to find every directory in a directory below baseoa, where oa config is}
   path := oaBasePath();

   {create a new mod list or clear existing one}
   modlist.Clear();

   {find first}
   result := FindFirst(UTF8ToSys(path)+'*', faReadOnly or faDirectory, src);
   if(result = 0) then begin
      repeat
         if(src.Attr and faDirectory > 0) then begin
            if(src.Name <> '.') and (src.Name <> '..') then begin
               cf := path+DirectorySeparator+src.Name+DirectorySeparator+oaCFG_FN;
               {only add potential mods if there seems to be a config file}
               if(FileExistsUTF8(cf)) then begin
                  modlist.Add(src.Name);
               end;
            end;
         end;
         result := FindNext(src);
      until (result <> 0);
   end;

   {we're done}
   FindClose(src);
end;

function oaHasMods(): boolean;
begin
   result := false;
   if(modlist.Count > 0) then result := true;
end;

function oaBasePath(): string;
begin
   result := ExtractFilePath(ExcludeTrailingPathDelimiter(ExtractFilePath(oacfgfile)));
end;

{ RENDERING }

procedure PreviewName(const thename: string; const shp: TShape);
var
   n, i, px, py: longint;
   c, nc: byte;
   tmpimg: TBGRABitmap;
   p: TBGRAPixel;
   pclr: longint = CLR_WHITE;
   name, normalname: string;

procedure setcolor();
begin
   p.red := colors[pclr][0];
   p.green := colors[pclr][1];
   p.blue := colors[pclr][2];
   p.alpha := 255;
end;

function normalchar(): boolean;
begin
   result := true;
   c := byte(name[i]);
   if(i < n) then
      nc := byte(name[i+1])
   else
      nc := 0;

   {check for a color character}
   if(c = byte('^')) then begin
      {we can only use colors from 48(0) to 56(8)}
      if(nc >= 48) and (nc <= 56) then begin
         pclr := nc - 48;

         setcolor();
         inc(i, 2); result := false;
      end;
   end;
end;

procedure getnormalname();
begin
   n := length(name);
   i := 1;
   if(n > 0) then repeat
      if(normalchar()) then begin
         normalname := normalname + name[i];
         inc(i);
      end;
   until (i > n);
end;

begin
   if(thename <> '') then begin
      name := thename;
      normalname := '';

      {get name}
      getnormalname();

      {calculate central positions}
      px := (shp.Width div 2) - (length(normalname)*CHAR_WIDTH div 2);
      py := shp.Height div 2 - (CHAR_HEIGHT div 2);

      {set initial color}
      pclr := CLR_WHITE; setcolor();

      if(n > 0) then begin
         {need a new temporary image}
         tmpimg := TBGRABitmap.Create(CHAR_WIDTH, CHAR_HEIGHT);

         {start going through the text}
         i := 1;
         repeat
            if(i > n) then break;

            {get the current and next characters}
            if(not normalchar()) then continue;

            {we can only draw a character if it exists in the image table}
            if(c <= MAX_CHAR) then begin
               {fill the temporary image with desired color and multiply}
               tmpimg.Fill(p);
               tmpimg.BlendImage(0, 0, imgs[c], boMultiply);

               {draw image}
               tmpimg.Draw(shp.Canvas, px, py, false);
            end;

            {move on to the next character}
            px := px + CHAR_WIDTH;
            inc(i);
         until (i > n);
      end;
   end;
end;

{ WINDOW PROPERTIES }
procedure LoadWndProps(const section: string; var props: TWindowProps);
begin
   if(iniCfg.SectionExists(section)) then begin
      props.loaded := true;

      props.w := iniCfg.ReadInteger(section, 'Width', -1);
      props.h := iniCfg.ReadInteger(section, 'Height', -1);
      props.x := iniCfg.ReadInteger(section, 'X', -1);
      props.y := iniCfg.ReadInteger(section, 'Y', -1);

      if((props.x = 0) or (props.y = 0)) then
         props.loaded := false;
   end;
end;

procedure WriteWndProps(const section: string; var props: TWindowProps);
begin
   iniCfg.WriteInteger(section, 'Width', props.w);
   iniCfg.WriteInteger(section, 'Height', props.h);
   iniCfg.WriteInteger(section, 'X', props.x);
   iniCfg.WriteInteger(section, 'Y', props.y);
end;

procedure ApplyWndProps(var props: TWindowProps; var frm: TForm);
begin
   if(props.loaded) then begin
      frm.Position := poDesigned;
      frm.Left := props.x;
      frm.Top := props.y;
      frm.Width := props.w;
      frm.Height := props.h;
   end;
end;

procedure StoreWndProps(var props: TWindowProps; var f: TForm);
begin
   props.w := f.Width;
   props.h := f.Height;
   props.x := f.Left;
   props.y := f.Top;
end;

{ URL ENCODING }

function urlEncode(const s: string): string;
var
  i: longint;

begin
  Result := '';
  for i := 1 to Length(s) do begin
    case s[i] of
      'A'..'Z', 'a'..'z', '0'..'9', '-', '_', '.': result := result + s[i];
      ' ': result := result + '%20';
      else
        result := result + '%' + IntToHex(Ord(s[i]), 2);
    end;
  end;
end;

function urlDecode(const s: string): string;
var
  i: Integer;
  hex: string;
  code: Integer;

begin
  result := '';
  i := 1;

   while i <= Length(s) do begin
      if(s[i] <> '%') then
         result := result + s[i]
      else begin
         if i <= (Length(s) - 2) then begin
            hex := s[i+1] + s[i+2];
            code := SysUtils.StrToIntDef('$' + hex, -1);
            inc(i, 2);
         end else
            code := -1;

         if code = -1 then begin
            raise SysUtils.EConvertError.Create('Invalid hex digit in URL');
         end;
         result := result + Chr(code);
      end;

      inc(i);
   end;
end;

INITIALIZATION
   modlist := TStringList.Create();
   modlistsel := TStringList.Create();

   oacfgfile := oaGetCfgFile();
   {$IFDEF WINDOWS}
   localCfgPath := SysToUTF8(ExcludeTrailingPathDelimiter(ExtractFilePath(ParamStr(0)))+DirectorySeparator);
   {$ENDIF}

END.
